import threading

from cupy_backends.cuda.api cimport runtime
from cupy_backends.cuda cimport stream as backends_stream
from cupy.cuda cimport graph

from cupy import _util


cdef object _thread_local = threading.local()


cdef class _ThreadLocal:
    # We keep both current_stream and current_stream_stack because the
    # former is also used when calling stream.use(). This bookkeeping enables
    # correct rewinding when "with" blocks are mixed with ".use()" (though
    # this is considered an anti-pattern).

    cdef list current_stream  # list of object
    cdef list current_device_id_stack  # list of int
    cdef list current_stream_stack  # list of list

    def __init__(self):
        cdef int i, num_devices = runtime.getDeviceCount()
        self.current_stream = []
        self.current_device_id_stack = []
        self.current_stream_stack = []
        for i in range(num_devices):
            default_stream = get_default_stream()
            self.current_stream.append(default_stream)
            self.current_stream_stack.append([default_stream])

    @staticmethod
    cdef _ThreadLocal get():
        try:
            tls = _thread_local.tls
        except AttributeError:
            tls = _thread_local.tls = _ThreadLocal()
        return <_ThreadLocal>tls

    cdef void push_stream(self, stream, int device_id) except*:
        assert device_id >= 0
        self.current_stream_stack[device_id].append(stream)
        # record device_id to prevent from popping the wrong stream at exit
        self.current_device_id_stack.append(device_id)
        self.set_current_stream(stream)

    cdef void pop_stream(self) except*:
        cdef int device_id = self.current_device_id_stack.pop()
        self.current_stream_stack[device_id].pop()
        prev_stream = self.current_stream_stack[device_id][-1]
        self.set_current_stream(prev_stream)
        assert len(self.current_stream_stack[device_id]) >= 1

    cdef set_current_stream(self, stream):
        cdef intptr_t ptr = <intptr_t>stream.ptr
        cdef int device_id = stream.device_id
        if device_id == -1:
            device_id = runtime.getDevice()
        backends_stream.set_current_stream_ptr(ptr, device_id)
        self.current_stream[device_id] = stream

    cdef get_current_stream(self, int device_id=-1):
        if device_id == -1:
            device_id = runtime.getDevice()
        stream_ref = self.current_stream[device_id]
        return stream_ref

    cdef intptr_t get_current_stream_ptr(self):
        return backends_stream.get_current_stream_ptr()


cdef get_default_stream():
    return Stream.ptds if backends_stream.is_ptds_enabled() else Stream.null


cdef intptr_t get_current_stream_ptr():
    """C API to get current CUDA stream pointer.

    Returns:
        intptr_t: The current CUDA stream pointer.
    """
    tls = _ThreadLocal.get()
    return <intptr_t>tls.get_current_stream_ptr()


cpdef get_current_stream(int device_id=-1):
    """Gets the current CUDA stream for the specified CUDA device.

    Args:
        device_id (int, optional): Index of the device to check for the current
            stream. The currently active device is selected by default.
    Returns:
        cupy.cuda.Stream: The current CUDA stream.
    """
    tls = _ThreadLocal.get()
    return tls.get_current_stream(device_id)


class Event(object):

    """CUDA event, a synchronization point of CUDA streams.

    This class handles the CUDA event handle in RAII way, i.e., when an Event
    instance is destroyed by the GC, its handle is also destroyed.

    Args:
        block (bool): If ``True``, the event blocks on the
            :meth:`~cupy.cuda.Event.synchronize` method.
        disable_timing (bool): If ``True``, the event does not prepare the
            timing data.
        interprocess (bool): If ``True``, the event can be passed to other
            processes.

    Attributes:
        ~Event.ptr (intptr_t): Raw event handle.

    """

    def __init__(self, block=False, disable_timing=False, interprocess=False):
        self.ptr = 0

        if interprocess and not disable_timing:
            raise ValueError('Timing must be disabled for interprocess events')
        flag = ((block and runtime.eventBlockingSync) |
                (disable_timing and runtime.eventDisableTiming) |
                (interprocess and runtime.eventInterprocess))
        self.ptr = runtime.eventCreateWithFlags(flag)

    def __del__(self, is_shutting_down=_util.is_shutting_down):
        if is_shutting_down():
            return
        if self.ptr:
            runtime.eventDestroy(self.ptr)

    @property
    def done(self):
        """True if the event is done."""
        return runtime.eventQuery(self.ptr) == 0  # cudaSuccess

    def record(self, stream=None):
        """Records the event to a stream.

        Args:
            stream (cupy.cuda.Stream): CUDA stream to record event. The null
                stream is used by default.

        .. seealso:: :meth:`cupy.cuda.Stream.record`

        """
        if stream is None:
            stream_ptr = get_current_stream_ptr()
        else:
            stream_ptr = stream.ptr
        runtime.eventRecord(self.ptr, stream_ptr)

    def synchronize(self):
        """Synchronizes all device work to the event.

        If the event is created as a blocking event, it also blocks the CPU
        thread until the event is done.

        """
        runtime.eventSynchronize(self.ptr)


def get_elapsed_time(start_event, end_event):
    """Gets the elapsed time between two events.

    Args:
        start_event (Event): Earlier event.
        end_event (Event): Later event.

    Returns:
        float: Elapsed time in milliseconds.

    """
    return runtime.eventElapsedTime(start_event.ptr, end_event.ptr)


cdef int check_stream_device_match(int device_id) except? -1:
    """Check if the stream was created on the current device."""
    cdef int curr_dev = runtime.getDevice()
    if device_id == -1:
        device_id = curr_dev
    if device_id != curr_dev:
        raise RuntimeError(
            f'The device that the stream is created on ({device_id})'
            f' does not match with the current device ({curr_dev})')
    return device_id


class _BaseStream:

    """CUDA stream.

    Attributes:
        ~Stream.ptr (intptr_t): Raw stream handle.
        ~Stream.device_id (int): The ID of the device that the stream was
            created on.

    """

    def __init__(self, ptr, device_id):
        self.ptr = ptr
        self.device_id = device_id
        self._graph_host_funcargs: list[tuple] | None = None

    def __eq__(self, other):
        # This operator needed as the ptr may be shared between multiple Stream
        # instances (e.g, `Stream.null` singleton and `Stream(null=True)` or
        # `ExternalStream`s).
        if isinstance(other, _BaseStream):
            return self.ptr == other.ptr
        return False

    def __enter__(self):
        # N.B. for maintainers: do not use this context manager in CuPy
        # codebase. See #5943 and #5963.
        tls = _ThreadLocal.get()
        cdef int device_id = self.device_id
        device_id = check_stream_device_match(device_id)
        tls.push_stream(self, device_id)
        return self

    def __exit__(self, *args):
        tls = _ThreadLocal.get()
        tls.pop_stream()

    def __repr__(self):
        return '<{} {} (device {})>'.format(
            type(self).__name__, self.ptr, self.device_id)

    def __hash__(self):
        return self.ptr

    def use(self):
        """Makes this stream current.

        If you want to switch a stream temporarily, use the *with* statement.
        """
        tls = _ThreadLocal.get()
        cdef int device_id = self.device_id
        check_stream_device_match(device_id)
        tls.set_current_stream(self)
        return self

    @property
    def done(self):
        """True if all work on this stream has been done."""
        return runtime.streamQuery(self.ptr) == 0  # cudaSuccess

    def synchronize(self):
        """Waits for the stream completing all queued work."""
        runtime.streamSynchronize(self.ptr)

    def add_callback(self, callback, arg):
        """Adds a callback that is called when all queued work is done.

        Args:
            callback (function): Callback function. It must take three
                arguments (Stream object, int error status, and user data
                object), and returns nothing.
            arg (object): Argument to the callback.

        .. note::
            Whenever possible, use the :meth:`launch_host_func` method
            instead of this one, as it may be deprecated and removed from
            CUDA at some point. Also note that this function does not support
            CUDA graph capture.

        """
        def f(stream, status, dummy):
            callback(self, status, arg)

        runtime.streamAddCallback(self.ptr, f, 0)

    def launch_host_func(self, callback, arg):
        """Launch a callback on host when all queued work is done.

        Args:
            callback (function): Callback function. It must take only one
                argument (user data object), and returns nothing.
            arg (object): Argument to the callback.

        .. seealso:: `cudaLaunchHostFunc()`_

        .. _cudaLaunchHostFunc():
            https://docs.nvidia.com/cuda/cuda-runtime-api/group__CUDART__EXECUTION.html#group__CUDART__EXECUTION_1g05841eaa5f90f27124241baafb3e856f

        """
        if self.is_capturing():
            func_arg = (callback, arg)
            runtime._launchHostFuncUnmanaged(self.ptr, func_arg)
            # Keep reference to the host callback function and arguments.
            # The ownership is later transferred to the captured Graph
            # instance in `end_capture()`.
            self._graph_host_funcargs.append(func_arg)
        else:
            def f(dummy):
                callback(arg)
            runtime.launchHostFunc(self.ptr, f, 0)

    def record(self, event=None):
        """Records an event on the stream.

        Args:
            event (None or cupy.cuda.Event): CUDA event. If ``None``, then a
                new plain event is created and used.

        Returns:
            cupy.cuda.Event: The recorded event.

        .. seealso:: :meth:`cupy.cuda.Event.record`

        """
        if event is None:
            event = Event()
        runtime.eventRecord(event.ptr, self.ptr)
        return event

    def wait_event(self, event):
        """Makes the stream wait for an event.

        The future work on this stream will be done after the event.

        Args:
            event (cupy.cuda.Event): CUDA event.

        """
        runtime.streamWaitEvent(self.ptr, event.ptr)

    def begin_capture(self, mode=None):
        """Begin stream capture to construct a CUDA graph.

        A call to this function must be paired with a call to
        :meth:`end_capture` to complete the capture.

        .. code-block:: python

            # create a non-blocking stream for the purpose of capturing
            s1 = cp.cuda.Stream(non_blocking=True)
            with s1:
                s1.begin_capture()
                # ... perform operations to construct a graph ...
                g = s1.end_capture()

            # the returned graph can be launched on any stream (including s1)
            g.launch(stream=s1)
            s1.synchronize()

            s2 = cp.cuda.Stream()
            with s2:
                g.launch()
            s2.synchronize()

        Args:
            mode (int): The stream capture mode. Default is
                :data:`~cupy.cuda.runtime.streamCaptureModeRelaxed`.

        .. note:: During the stream capture, synchronous device-host transfers
            are not allowed. This has a particular implication for CuPy APIs,
            as some functions that internally require synchronous transfer
            would not work as expected and an exception would be raised. For
            further constraints of CUDA stream capture, please refer to the
            CUDA Programming Guide.

        .. note:: Currently this capability is not supported on HIP.

        .. seealso:: `cudaStreamBeginCapture()`_

        .. _cudaStreamBeginCapture():
            https://docs.nvidia.com/cuda/cuda-runtime-api/group__CUDART__STREAM.html#group__CUDART__STREAM_1g793d7d4e474388ddfda531603dc34aa3

        """
        if runtime._is_hip_environment:
            raise RuntimeError('This function is not supported on HIP')
        if self.ptr == 0 or self.ptr == 1:
            raise RuntimeError('cannot capture on the default (legacy) stream')
        if mode is None:
            # We default to the relaxed mode for the following reason: During
            # the capture the memory pool might need to increase size. If it's
            # backed by cudaMalloc, which is an "unsafe" API, other modes would
            # not permit such a (legitimate) operation. See the API doc for
            # "cudaThreadExchangeStreamCaptureMode" for further detail.
            # (Though, ideally stream capture should be used together with
            # the async APIs, such as cudaMallocAsync.)
            mode = runtime.streamCaptureModeRelaxed
        runtime.streamBeginCapture(self.ptr, mode)
        self._graph_host_funcargs = []

    def end_capture(self):
        """End stream capture and retrieve the constructed CUDA graph.

        Returns:
            cupy.cuda.Graph:
                A CUDA graph object that encapsulates the captured work.

        .. note:: Currently this capability is not supported on HIP.

        .. seealso:: `cudaStreamEndCapture()`_

        .. _cudaStreamEndCapture():
            https://docs.nvidia.com/cuda/cuda-runtime-api/group__CUDART__STREAM.html#group__CUDART__STREAM_1gf5a0efebc818054ceecd1e3e5e76d93e

        """
        if runtime._is_hip_environment:
            raise RuntimeError('This function is not supported on HIP')
        host_funcargs = self._graph_host_funcargs
        self._graph_host_funcargs = None
        cdef intptr_t g = runtime.streamEndCapture(self.ptr)
        return graph.Graph.from_stream(g, host_funcargs)

    def is_capturing(self):
        """Check if the stream is capturing.

        Returns:
            bool:
                If the capturing status is successfully queried, the returned
                value indicates the capturing status. An exception could be
                raised if such a query is illegal, please refer to the CUDA
                Programming Guide for detail.

        """
        # TODO(leofang): is it better to be a property?
        if runtime._is_hip_environment:
            raise RuntimeError('This function is not supported on HIP')
        try:
            return runtime.streamIsCapturing(self.ptr)
        except RuntimeError:  # can be RuntimeError or CUDARuntimeError
            raise


class Stream(_BaseStream):

    """CUDA stream.

    This class handles the CUDA stream handle in RAII way, i.e., when an Stream
    instance is destroyed by the GC, its handle is also destroyed.

    Note that if both ``null`` and ``ptds`` are ``False``, a plain new
    stream is created.

    Args:
        null (bool): If ``True``, the stream is a null stream (i.e. the default
            stream that synchronizes with all streams). Note that you can also
            use the ``Stream.null`` singleton object instead of creating a new
            null stream object.
        ptds (bool): If ``True`` and ``null`` is ``False``, the per-thread
            default stream is used. Note that you can also use the
            ``Stream.ptds`` singleton object instead of creating a new
            per-thread default stream object.
        non_blocking (bool): If ``True`` and both ``null`` and ``ptds`` are
            ``False``, the stream does not synchronize with the NULL stream.

    Attributes:
        ~Stream.ptr (intptr_t): Raw stream handle.
        ~Stream.device_id (int): The ID of the device that the stream was
            created on. The value ``-1`` is used for the singleton stream
            objects.

    """

    null = None
    ptds = None

    def __init__(self, null=False, non_blocking=False, ptds=False):
        if null:
            # TODO(pentschev): move to streamLegacy. This wasn't possible
            # because of a NCCL bug that should be fixed in the version
            # following 2.8.3-1.
            ptr = 0
            device_id = -1
        elif ptds:
            if runtime._is_hip_environment:
                raise ValueError('HIP does not support per-thread '
                                 'default stream (ptds)')
            ptr = runtime.streamPerThread
            device_id = -1
        elif non_blocking:
            ptr = runtime.streamCreateWithFlags(runtime.streamNonBlocking)
            device_id = runtime.getDevice()
        else:
            ptr = runtime.streamCreate()
            device_id = runtime.getDevice()
        super().__init__(ptr, device_id)

    def __del__(self, is_shutting_down=_util.is_shutting_down):
        if is_shutting_down():
            return
        if self.ptr not in (0, runtime.streamLegacy, runtime.streamPerThread):
            runtime.streamDestroy(self.ptr)
        # Note that we can not release memory pool of the stream held in CPU
        # because the memory would still be used in kernels executed in GPU.


class ExternalStream(_BaseStream):

    """CUDA stream not managed by CuPy.

    This class allows to use external streams in CuPy by providing the
    stream pointer obtained from the CUDA runtime call.
    The user is in charge of managing the life-cycle of the stream.

    Args:
        ptr (intptr_t): Address of the `cudaStream_t` object.
        device_id (int): The ID of the device that the stream was created on.
            Default is ``-1``, indicating it is unknown.

    Attributes:
        ~Stream.ptr (intptr_t): Raw stream handle.
        ~Stream.device_id (int): The ID of the device that the stream was
            created on. The value ``-1`` is used to indicate it is unknown.

    .. warning::
        If ``device_id`` is not specified, the user is required to ensure legal
        operations of the stream. Specifically, the stream must be used on the
        device that it was created on.

    """

    def __init__(self, ptr, device_id=-1):
        # It is in theory unsafe to just call runtime.getDevice() here, as the
        # stream pointer could come from a different device (although
        # unlikely). While we could use driver API combos cuStreamGetCtx ->
        # cuCtxSetCurrent -> cuCtxGetDevice -> ... to retrieve the device ID
        # associated with the stream, it is way too complicated and does not
        # work with HIP. Let us keep this as thin as possible.
        super().__init__(ptr, device_id)


Stream.null = Stream(null=True)
if not runtime._is_hip_environment:
    Stream.ptds = Stream(ptds=True)



from cupy._core._carray cimport shape_t
from cupy._core cimport _kernel
from cupy._core.core cimport _ndarray_base
from cupy.cuda cimport function


cdef Py_ssize_t _block_size


cpdef tuple _get_axis(object axis, Py_ssize_t ndim)

cpdef shape_t _get_out_shape(
    const shape_t& shape, tuple reduce_axis, tuple out_axis, bint keepdims)


cdef class _AbstractReductionKernel:

    cdef:
        readonly str name
        public str identity
        readonly tuple in_params
        readonly tuple out_params
        readonly tuple _params
        readonly str __name__
        readonly dict _cached_codes

    cpdef _ndarray_base _call(
        self,
        list in_args, list out_args,
        const shape_t& a_shape, axis, dtype,
        bint keepdims, bint reduce_dims, int device_id,
        stream, bint try_use_cub=*, bint sort_reduce_axis=*)

    cdef void _launch(
        self, out_block_num, block_size, block_stride,
        in_args, out_args, in_shape, out_shape, types,
        map_expr, reduce_expr, post_map_expr, reduce_type,
        stream, params)

    cdef tuple _get_expressions_and_types(
        self, list in_args, list out_args, dtype)

    cdef list _get_out_args(
        self, list out_args, tuple out_types, const shape_t& out_shape)

    cdef function.Function _get_function(
        self,
        tuple params, tuple arginfos, _kernel._TypeMap types,
        str map_expr, str reduce_expr, str post_map_expr, str reduce_type,
        Py_ssize_t block_size)


cdef class ReductionKernel(_AbstractReductionKernel):

    cdef:
        readonly int nin
        readonly int nout
        readonly int nargs
        readonly tuple params
        readonly str reduce_expr
        readonly str map_expr
        readonly str post_map_expr
        readonly object options
        readonly bint reduce_dims
        readonly object reduce_type
        readonly str preamble


cdef shape_t _set_permuted_args(
    list args, tuple axis_permutes, const shape_t& shape, tuple params)

cdef tuple _get_shape_and_strides(list in_args, list out_args)

cdef _optimizer_copy_arg(a)

cpdef create_reduction_func(
    name, ops, routine=*, identity=*, preamble=*, sort_reduce_axis=*)



import numpy
import pytest

import cupy
from cupy import testing


class C(cupy.ndarray):

    def __new__(cls, *args, info=None, **kwargs):
        obj = super().__new__(cls, *args, **kwargs)
        obj.info = info
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.info = getattr(obj, 'info', None)


class TestArrayUfunc:

    @testing.for_all_dtypes()
    def test_unary_op(self, dtype):
        a = cupy.array(numpy.array([0, 1, 2]), dtype=dtype)
        outa = numpy.sin(a)
        # numpy operation produced a cupy array
        assert isinstance(outa, cupy.ndarray)
        b = a.get()
        outb = numpy.sin(b)
        assert numpy.allclose(outa.get(), outb)

    @testing.for_all_dtypes()
    def test_unary_op_out(self, dtype):
        a = cupy.array(numpy.array([0, 1, 2]), dtype=dtype)
        b = a.get()
        outb = numpy.sin(b)
        # pre-make output with same type as input
        outa = cupy.array(numpy.array([0, 1, 2]), dtype=outb.dtype)
        numpy.sin(a, out=outa)
        assert numpy.allclose(outa.get(), outb)

    @testing.for_all_dtypes()
    def test_binary_op(self, dtype):
        a1 = cupy.array(numpy.array([0, 1, 2]), dtype=dtype)
        a2 = cupy.array(numpy.array([0, 1, 2]), dtype=dtype)
        outa = numpy.add(a1, a2)
        # numpy operation produced a cupy array
        assert isinstance(outa, cupy.ndarray)
        b1 = a1.get()
        b2 = a2.get()
        outb = numpy.add(b1, b2)
        assert numpy.allclose(outa.get(), outb)

    @testing.for_all_dtypes()
    def test_binary_op_out(self, dtype):
        a1 = cupy.array(numpy.array([0, 1, 2]), dtype=dtype)
        a2 = cupy.array(numpy.array([0, 1, 2]), dtype=dtype)
        outa = cupy.array(numpy.array([0, 1, 2]), dtype=dtype)
        numpy.add(a1, a2, out=outa)
        b1 = a1.get()
        b2 = a2.get()
        outb = numpy.add(b1, b2)
        assert numpy.allclose(outa.get(), outb)

    @testing.for_all_dtypes()
    def test_binary_mixed_op(self, dtype):
        a1 = cupy.array(numpy.array([0, 1, 2]), dtype=dtype)
        a2 = cupy.array(numpy.array([0, 1, 2]), dtype=dtype).get()
        with pytest.raises(TypeError):
            # attempt to add cupy and numpy arrays
            numpy.add(a1, a2)
        with pytest.raises(TypeError):
            # check reverse order
            numpy.add(a2, a1)
        with pytest.raises(TypeError):
            # reject numpy output from cupy
            numpy.add(a1, a1, out=a2)
        with pytest.raises(TypeError):
            # reject cupy output from numpy
            numpy.add(a2, a2, out=a1)
        with pytest.raises(ValueError):
            # bad form for out=
            # this is also an error with numpy array
            numpy.sin(a1, out=())
        with pytest.raises(ValueError):
            # bad form for out=
            # this is also an error with numpy array
            numpy.sin(a1, out=(a1, a1))

    @testing.numpy_cupy_array_equal()
    def test_indexing(self, xp):
        a = cupy.testing.shaped_arange((3, 1), xp)[:, :, None]
        b = cupy.testing.shaped_arange((3, 2), xp)[:, None, :]
        return a * b

    @testing.numpy_cupy_array_equal()
    def test_shares_memory(self, xp):
        a = cupy.testing.shaped_arange((1000, 1000), xp, 'int64')
        b = xp.transpose(a)
        a += b
        return a

    def test_subclass_unary_op(self):
        a = cupy.array([0, 1, 2]).view(C)
        a.info = 1
        outa = cupy.sin(a)
        assert isinstance(outa, C)
        assert outa.info is not None and outa.info == 1

        b = a.get()
        outb = numpy.sin(b)
        testing.assert_allclose(outa, outb)

    def test_subclass_binary_op(self):
        a0 = cupy.array([0, 1, 2]).view(C)
        a0.info = 1
        a1 = cupy.array([3, 4, 5]).view(C)
        a1.info = 2
        outa = cupy.add(a0, a1)
        assert isinstance(outa, C)
        # a0 is used to initialize outa.info
        assert outa.info is not None and outa.info == 1

        b0 = a0.get()
        b1 = a1.get()
        outb = numpy.add(b0, b1)
        testing.assert_allclose(outa, outb)

    def test_subclass_binary_op_mixed(self):
        a0 = cupy.array([0, 1, 2])
        a1 = cupy.array([3, 4, 5]).view(C)
        a1.info = 1
        outa = cupy.add(a0, a1)
        assert isinstance(outa, C)
        # The first appearance of C's instance is used to initialize outa.info
        assert outa.info is not None and outa.info == 1

        b0 = a0.get()
        b1 = a1.get()
        outb = numpy.add(b0, b1)
        testing.assert_allclose(outa, outb)

    @testing.numpy_cupy_array_equal()
    def test_ufunc_outer(self, xp):
        a = cupy.testing.shaped_arange((3, 4), xp)
        b = cupy.testing.shaped_arange((5, 6), xp)
        return numpy.add.outer(a, b)

    @testing.numpy_cupy_array_equal()
    def test_ufunc_at(self, xp):
        a = cupy.testing.shaped_arange((10,), xp)
        b = cupy.testing.shaped_arange((5,), xp)
        indices = xp.array([0, 3, 6, 7, 9])
        numpy.add.at(a, indices, b)
        return a

    @testing.numpy_cupy_array_equal()
    def test_ufunc_at_scalar(self, xp):
        a = cupy.testing.shaped_arange((10,), xp)
        b = 7
        indices = xp.array([0, 3, 6, 7, 9])
        numpy.add.at(a, indices, b)
        return a

    @testing.numpy_cupy_array_equal()
    def test_ufunc_reduce(self, xp):
        a = cupy.testing.shaped_arange((10, 12), xp)
        return numpy.add.reduce(a, axis=-1)

    @testing.numpy_cupy_array_equal()
    def test_ufunc_accumulate(self, xp):
        a = cupy.testing.shaped_arange((10, 12), xp)
        return numpy.add.accumulate(a, axis=-1)

    @testing.numpy_cupy_array_equal()
    def test_ufunc_reduceat(self, xp):
        a = cupy.testing.shaped_arange((10, 12), xp)
        indices = xp.array([0, 3, 6, 7, 9])
        return numpy.add.reduceat(a, indices, axis=-1)


class TestUfunc:
    @pytest.mark.parametrize('ufunc', [
        'add',
        'sin',
    ])
    @testing.numpy_cupy_equal()
    def test_types(self, xp, ufunc):
        types = getattr(xp, ufunc).types
        if xp == numpy:
            assert isinstance(types, list)
            types = list(dict.fromkeys(  # remove dups: numpy/numpy#7897
                sig for sig in types
                # CuPy does not support the following dtypes:
                # (c)longdouble, datetime, timedelta, and object.
                if not any(t in sig for t in 'GgMmO')
            ))
        return types

    @testing.numpy_cupy_allclose()
    def test_unary_out_tuple(self, xp):
        dtype = xp.float64
        a = testing.shaped_arange((2, 3), xp, dtype)
        out = xp.zeros((2, 3), dtype)
        ret = xp.sin(a, out=(out,))
        assert ret is out
        return ret

    @testing.numpy_cupy_allclose()
    def test_unary_out_positional_none(self, xp):
        dtype = xp.float64
        a = testing.shaped_arange((2, 3), xp, dtype)
        return xp.sin(a, None)

    @testing.numpy_cupy_allclose()
    def test_binary_out_tuple(self, xp):
        dtype = xp.float64
        a = testing.shaped_arange((2, 3), xp, dtype)
        b = xp.ones((2, 3), dtype)
        out = xp.zeros((2, 3), dtype)
        ret = xp.add(a, b, out=(out,))
        assert ret is out
        return ret

    @testing.numpy_cupy_allclose()
    def test_biary_out_positional_none(self, xp):
        dtype = xp.float64
        a = testing.shaped_arange((2, 3), xp, dtype)
        b = xp.ones((2, 3), dtype)
        return xp.add(a, b, None)

    @testing.numpy_cupy_allclose()
    def test_divmod_out_tuple(self, xp):
        dtype = xp.float64
        a = testing.shaped_arange((2, 3), xp, dtype)
        b = testing.shaped_reverse_arange((2, 3), xp, dtype)
        out0 = xp.zeros((2, 3), dtype)
        out1 = xp.zeros((2, 3), dtype)
        ret = xp.divmod(a, b, out=(out0, out1))
        assert ret[0] is out0
        assert ret[1] is out1
        return ret

    @testing.numpy_cupy_allclose()
    def test_divmod_out_positional_none(self, xp):
        dtype = xp.float64
        a = testing.shaped_arange((2, 3), xp, dtype)
        b = xp.ones((2, 3), dtype)
        return xp.divmod(a, b, None, None)

    @testing.numpy_cupy_allclose()
    def test_divmod_out_partial(self, xp):
        dtype = xp.float64
        a = testing.shaped_arange((2, 3), xp, dtype)
        b = testing.shaped_reverse_arange((2, 3), xp, dtype)
        out0 = xp.zeros((2, 3), dtype)
        ret = xp.divmod(a, b, out0)  # out1 is None
        assert ret[0] is out0
        return ret

    @testing.numpy_cupy_allclose()
    def test_divmod_out_partial_tuple(self, xp):
        dtype = xp.float64
        a = testing.shaped_arange((2, 3), xp, dtype)
        b = testing.shaped_reverse_arange((2, 3), xp, dtype)
        out1 = xp.zeros((2, 3), dtype)
        ret = xp.divmod(a, b, out=(None, out1))
        assert ret[1] is out1
        return ret



from itertools import chain
from typing import Any, Callable, Iterable, Optional

import numpy
from numpy.typing import ArrayLike
from numpy.typing import DTypeLike

import cupy
from cupy._core.core import ndarray
import cupy._creation.from_data as _creation_from_data
import cupy._core._routines_math as _math
import cupy._core._routines_statistics as _statistics
from cupy.cuda.device import Device
from cupy.cuda.stream import Stream
from cupy.cuda.stream import get_current_stream

from cupyx.distributed.array import _chunk
from cupyx.distributed.array._chunk import _Chunk
from cupyx.distributed.array import _data_transfer
from cupyx.distributed.array._data_transfer import _Communicator
from cupyx.distributed.array import _elementwise
from cupyx.distributed.array import _index_arith
from cupyx.distributed.array import _modes
from cupyx.distributed.array import _reduction
from cupyx.distributed.array import _linalg


class _MultiDeviceDummyMemory(cupy.cuda.memory.Memory):
    pass


class _MultiDeviceDummyPointer(cupy.cuda.memory.MemoryPointer):
    @property
    def device(self) -> Device:
        # This override is needed to assign an invalid device id
        # Since the array is not residing in a single device now
        return Device(-1)


def _make_chunk_async(src_dev, dst_dev, idx, src_array, comms):
    src_stream = get_current_stream(src_dev)
    with src_array.device:
        src_array = _creation_from_data.ascontiguousarray(src_array)
        src_data = _data_transfer._AsyncData(
            src_array, src_stream.record(), prevent_gc=src_array)
    with Device(dst_dev):
        dst_stream = get_current_stream()
        copied = _data_transfer._transfer(
            comms[src_dev], src_stream, src_data,
            comms[dst_dev], dst_stream, dst_dev)
        return _Chunk(copied.array, copied.ready, idx,
                      prevent_gc=src_data)


def _make_chunk_sync(src_dev, dst_dev, idx, src_array, comms):
    with Device(dst_dev):
        stream = get_current_stream()
        copied = _creation_from_data.array(src_array)
        return _Chunk(copied, stream.record(), idx, prevent_gc=src_array)


class DistributedArray(ndarray):
    """
    __init__(self, shape, dtype, chunks_map, mode=REPLICA, comms=None)

    Multi-dimensional array distributed across multiple CUDA devices.

    This class implements some elementary operations that :class:`cupy.ndarray`
    provides. The array content is split into chunks, contiguous arrays
    corresponding to slices of the original array. Note that one device can
    hold multiple chunks.

    This direct constructor is designed for internal calls. Users should create
    distributed arrays using :func:`distributed_array`.

    Args:
        shape (tuple of ints): Shape of created array.
        dtype (dtype_like): Any object that can be interpreted as a numpy data
            type.
        chunks_map (dict from int to list of chunks): Lists of chunk objects
            associated with each device.
        mode (mode object, optional): Mode that determines how overlaps
            of the chunks are interpreted. Defaults to
            ``cupyx.distributed.array.REPLICA``.
        comms (optional): Communicator objects which a distributed array
            hold internally. Sharing them with other distributed arrays can
            save time because their initialization is a costly operation.

    .. seealso::
            :attr:`DistributedArray.mode` for details about modes.
    """

    _chunks_map: dict[int, list[_Chunk]]
    _mode: _modes.Mode
    _streams: dict[int, Stream]
    _comms: dict[int, _Communicator]

    def __new__(
        cls, shape: tuple[int, ...], dtype: DTypeLike,
        chunks_map: dict[int, list[_Chunk]],
        mode: _modes.Mode = _modes.REPLICA,
        comms: Optional[dict[int, _Communicator]] = None,
    ) -> 'DistributedArray':
        mem = _MultiDeviceDummyMemory(0)
        memptr = _MultiDeviceDummyPointer(mem, 0)
        obj = super().__new__(cls, shape, dtype, memptr=memptr)
        obj._chunks_map = chunks_map

        obj._mode = mode

        obj._streams = {}
        obj._comms = comms if comms is not None else {}

        return obj

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def __array_finalize__(self, obj):
        if obj is not None:
            raise RuntimeError(
                'Distributed array can only be instantiated by an explicit'
                'constructor call')

    @property
    def mode(self) -> _modes.Mode:
        """Describe how overlaps of the chunks are interpreted.

        In the replica mode, chunks are guaranteed to have identical values on
        their overlapping segments. In other modes, they are not necessarily
        identical and represent the original data as their max, sum, etc.

        :class:`DistributedArray` currently supports
        ``cupyx.distributed.array.REPLICA``, ``cupyx.distributed.array.MIN``,
        ``cupyx.distributed.array.MAX``, ``cupyx.distributed.array.SUM``,
        ``cupyx.distributed.array.PROD`` modes.

        Many operations on distributed arrays including :class:`cupy.ufunc`
        and :func:`~cupyx.distributed.array.matmul` involve changing their mode
        beforehand. These mode conversions are done automatically, so in most
        cases users do not have to manage modes manually.

        Example:
            >>> A = distributed_array(
            ...     cupy.arange(6).reshape(2, 3),
            ...     make_2d_index_map([0, 2], [0, 1, 3],
            ...                       [[{0}, {1, 2}]]))
            >>> B = distributed_array(
            ...     cupy.arange(12).reshape(3, 4),
            ...     make_2d_index_map([0, 1, 3], [0, 2, 4],
            ...                       [[{0}, {0}],
            ...                        [{1}, {2}]]))
            >>> C = A @ B
            >>> C
            array([[20, 23, 26, 29],
                   [56, 68, 80, 92]])
            >>> C.mode
            'sum'
            >>> C.all_chunks()
            {0: [array([[0, 0],
                        [0, 3]]),     # left half
                 array([[0, 0],
                        [6, 9]])],    # right half
             1: [array([[20, 23],
                        [56, 65]])],  # left half
             2: [array([[26, 29],
                        [74, 83]])]}  # right half
            >>> C_replica = C.change_mode('replica')
            >>> C_replica.mode
            'replica'
            >>> C_replica.all_chunks()
            {0: [array([[20, 23],
                        [56, 68]]),   # left half
                 array([[26, 29],
                        [80, 92]])],  # right half
             1: [array([[20, 23],
                        [56, 68]])],  # left half
             2: [array([[26, 29],
                        [80, 92]])]}  # right half
        """
        return self._mode

    @property
    def devices(self) -> Iterable[int]:
        """A collection of device IDs holding part of the data."""
        return self._chunks_map.keys()

    @property
    def index_map(self) -> dict[int, list[tuple[slice, ...]]]:
        """Indices for the chunks that devices with designated IDs own."""
        return {dev: [chunk.index for chunk in chunks]
                for dev, chunks in self._chunks_map.items()}

    def all_chunks(self) -> dict[int, list[ndarray]]:
        """Return the chunks with all buffered data flushed.

        Buffered data are created in situations such as resharding and mode
        changing.
        """
        chunks_map: dict[int, list[ndarray]] = {}
        for dev, chunks in self._chunks_map.items():
            chunks_map[dev] = []
            for chunk in chunks:
                chunk.flush(self._mode)
                chunks_map[dev].append(chunk.array)
        return chunks_map

    def _prepare_comms_and_streams(self, devices: Iterable[int]) -> None:
        # Ensure communicators and streams are prepared for communication
        # between `devices` and the devices currently owning chunks
        devices = self._chunks_map.keys() | devices

        if not devices.issubset(self._comms.keys()):
            self._comms = _data_transfer._create_communicators(devices)

        for dev in devices - self._streams.keys():
            with Device(dev):
                self._streams[dev] = Stream()

    def __cupy_override_elementwise_kernel__(self, kernel, *args, **kwargs):
        # This method is called from cupy.ufunc and cupy.ElementwiseKernel
        # to dispatch elementwise operations
        return _elementwise._execute(kernel, args, kwargs)

    def __cupy_override_reduction_kernel__(
            self, kernel, axis, dtype, out, keepdims) -> Any:
        # This method is called from _SimpleReductionKernel and elementary
        # reduction methods of ndarray to dispatch reduction operations
        # TODO: Support user-defined ReductionKernel
        if axis is None:
            raise RuntimeError('axis must be specified')
        if out is not None:
            raise RuntimeError('Argument `out` is not supported')
        if keepdims:
            raise RuntimeError('Argument `keepdims` is not supported')

        return _reduction._execute(self, kernel, axis, dtype)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        if ufunc.__name__ == 'matmul' and method == '__call__':
            return _linalg.matmul(*inputs, **kwargs)
        return NotImplemented

    def __matmul__(x, y):
        if isinstance(y, DistributedArray):
            return _linalg.matmul(x, y)
        else:
            return NotImplemented

    def _copy_chunks_map_in_replica_mode(self) -> dict[int, list[_Chunk]]:
        # Return a copy of self.chunks_map in the replica mode
        chunks_map = {}
        for dev, chunks in self._chunks_map.items():
            chunks_map[dev] = [chunk.copy() for chunk in chunks]

        if self._mode is not _modes.REPLICA:
            self._prepare_comms_and_streams(self._chunks_map.keys())
            _chunk._all_reduce_intersections(
                self._mode, self.shape, chunks_map, self._comms, self._streams)

        return chunks_map

    def _copy_chunks_map_in_op_mode(
            self, op_mode: _modes._OpMode) -> dict[int, list[_Chunk]]:
        # Return a copy of self.chunks_map in the given op mode
        chunks_map = self._copy_chunks_map_in_replica_mode()

        for chunk in chain.from_iterable(chunks_map.values()):
            chunk.flush(_modes.REPLICA)

        chunks_list = list(chain.from_iterable(chunks_map.values()))
        identity = op_mode.identity_of(self.dtype)

        # TODO: Fair distribution of work
        # In the current implementation, devices that appear earlier have to
        # execute set_identity_on_intersection repeatedly, whereas the last
        # device has no work to do
        for i in range(len(chunks_list)):
            a_chunk = chunks_list[i]
            for j in range(i + 1, len(chunks_list)):
                b_chunk = chunks_list[j]
                a_chunk.set_identity_on_intersection(
                    b_chunk.index, self.shape, identity)

        return chunks_map

    def _to_op_mode(self, op_mode: _modes.Mode) -> 'DistributedArray':
        # Return a view or a copy of the chunks_map in the given mode
        if self._mode is op_mode:
            return self

        if len(self._chunks_map) == 1:
            chunks, = self._chunks_map.values()
            if len(chunks) == 1:
                chunks[0].flush(self._mode)
                return DistributedArray(
                    self.shape, self.dtype, self._chunks_map,
                    op_mode, self._comms)
        if op_mode is _modes.REPLICA:
            chunks_map = self._copy_chunks_map_in_replica_mode()
        else:
            assert op_mode is not None
            chunks_map = self._copy_chunks_map_in_op_mode(op_mode)
        return DistributedArray(
            self.shape, self.dtype, chunks_map, op_mode, self._comms)

    def change_mode(self, mode: _modes.Mode) -> 'DistributedArray':
        """Return a view or a copy in the given mode.

        Args:
            mode (mode Object): How overlaps of
                the chunks are interpreted.

        .. seealso::
                :attr:`DistributedArray.mode` for details about modes.
        """
        return self._to_op_mode(mode)

    def reshard(self, index_map: dict[int, Any]) -> 'DistributedArray':
        """Return a view or a copy having the given index_map.

        Data transfers across devices are done on separate streams created
        internally. To make them asynchronous, transferred data is buffered and
        reflected to the chunks when necessary.

        Args:
            index_map (dict from int to array indices): Indices for the chunks
                that devices with designated IDs own. The current index_map of
                a distributed array can be obtained from
                :attr:`DistributedArray.index_map`.
        """
        new_index_map = _index_arith._normalize_index_map(
            self.shape, index_map)
        if new_index_map == self.index_map:
            return self

        old_chunks_map = self._chunks_map
        new_chunks_map: dict[int, list[_Chunk]] = {}

        # Set up new_chunks_map compatible with new_index_map
        # as placeholders of chunks
        for dev, idxs in new_index_map.items():
            new_chunks_map[dev] = []

            for idx in idxs:
                with Device(dev):
                    dst_shape = _index_arith._shape_after_indexing(
                        self.shape, idx)
                    new_chunk = _Chunk.create_placeholder(dst_shape, dev, idx)
                    new_chunks_map[dev].append(new_chunk)

        self._prepare_comms_and_streams(index_map.keys())

        # Data transfer from old chunks to new chunks
        # TODO: Reorder transfers to minimize latency

        # The current implementation transfers the same data multiple times
        # where chunks overlap. This is particularly problematic when matrix
        # multiplication is involved, where one block tends to be shared
        # between multiple devices
        # TODO: Avoid duplicate data transfers
        for src_chunk in chain.from_iterable(old_chunks_map.values()):
            src_chunk.flush(self._mode)

            if self._mode is not _modes.REPLICA:
                src_chunk = src_chunk.copy()

            for dst_chunk in chain.from_iterable(new_chunks_map.values()):
                src_chunk.apply_to(
                    dst_chunk, self._mode, self.shape,
                    self._comms, self._streams)

        return DistributedArray(
            self.shape, self.dtype, new_chunks_map, self._mode, self._comms)

    def get(
        self, stream=None, order='C', out=None, blocking=True
    ) -> numpy.ndarray:
        """Return a copy of the array on the host memory."""
        if stream is not None:
            raise RuntimeError('Argument `stream` not supported')
        if order != 'C':
            raise RuntimeError('Argument `order` not supported')
        if out is not None:
            raise RuntimeError('Argument `out` not supported')

        for chunk in chain.from_iterable(self._chunks_map.values()):
            chunk.flush(self._mode)

        if self._mode is _modes.REPLICA:
            np_array = numpy.empty(self.shape, dtype=self.dtype)
        else:
            identity = self._mode.identity_of(self.dtype)
            np_array = numpy.full(self.shape, identity, self.dtype)

        # We avoid 0D array because we expect data[idx] to return a view
        np_array = numpy.atleast_1d(np_array)

        for chunk in chain.from_iterable(self._chunks_map.values()):
            chunk.ready.synchronize()
            idx = chunk.index
            if self._mode is _modes.REPLICA:
                np_array[idx] = cupy.asnumpy(chunk.array)
            else:
                self._mode.numpy_func(
                    np_array[idx], cupy.asnumpy(chunk.array), np_array[idx])

        # Undo numpy.atleast_1d
        return np_array.reshape(self.shape)

    # -----------------------------------------------------
    # Overriding unsupported methods inherited from ndarray
    # -----------------------------------------------------

    def __getitem__(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support __getitem__.')

    def __setitem__(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support __setitem__.')

    def __len__(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support __len__.')

    def __iter__(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support __iter__.')

    def __copy__(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support __copy__.')

    def all(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support all.')

    def any(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support any.')

    def argmax(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support argmax.')

    def argmin(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support argmin.')

    def argpartition(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support argpartition.')

    def argsort(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support argsort.')

    def astype(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support astype.')

    def choose(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support choose.')

    def clip(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support clip.')

    def compress(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support compress.')

    def copy(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support copy.')

    def cumprod(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support cumprod.')

    def cumsum(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support cumsum.')

    def diagonal(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support diagonal.')

    def dot(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support dot.')

    def dump(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support dump.')

    def dumps(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support dumps.')

    def fill(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support fill.')

    def flatten(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support flatten.')

    def item(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support item.')

    def max(self, axis=None, out=None, keepdims=False):
        """Return the maximum along a given axis.

        .. note::

            Currently, it only supports non-``None`` values for ``axis`` and
            the default values for ``out`` and ``keepdims``.

        .. seealso::
           :meth:`cupy.ndarray.max`, :meth:`numpy.ndarray.max`
        """
        return self.__cupy_override_reduction_kernel__(
            _statistics.amax, axis, None, out, keepdims)

    def mean(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support mean.')

    def min(self, axis=None, out=None, keepdims=False):
        """Return the minimum along a given axis.

        .. note::

            Currently, it only supports non-``None`` values for ``axis`` and
            the default values for ``out`` and ``keepdims``.

        .. seealso::
           :meth:`cupy.ndarray.min`, :meth:`numpy.ndarray.min`
        """
        return self.__cupy_override_reduction_kernel__(
            _statistics.amin, axis, None, out, keepdims)

    def nonzero(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support nonzero.')

    def partition(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support partition.')

    def prod(self, axis=None, dtype=None, out=None, keepdims=None):
        """Return the minimum along a given axis.

        .. note::

            Currently, it only supports non-``None`` values for ``axis`` and
            the default values for ``out`` and ``keepdims``.

        .. seealso::
           :meth:`cupy.ndarray.prod`, :meth:`numpy.ndarray.prod`
        """
        if dtype is None:
            return self.__cupy_override_reduction_kernel__(
                _math.prod_auto_dtype, axis, dtype, out, keepdims)
        else:
            return self.__cupy_override_reduction_kernel__(
                _math.prod_keep_dtype, axis, dtype, out, keepdims)

    def ptp(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support ptp.')

    def put(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support put.')

    def ravel(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support ravel.')

    def reduced_view(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support reduced_view.')

    def repeat(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support repeat.')

    def reshape(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support reshape.')

    def round(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support round.')

    def scatter_add(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support scatter_add.')

    def scatter_max(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support scatter_max.')

    def scatter_min(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support scatter_min.')

    def searchsorted(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support searchsorted.')

    def set(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support set.')

    def sort(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support sort.')

    def squeeze(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support squeeze.')

    def std(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support std.')

    def sum(self, axis=None, dtype=None, out=None, keepdims=False):
        """Return the minimum along a given axis.

        .. note::

            Currently, it only supports non-``None`` values for ``axis`` and
            the default values for ``out`` and ``keepdims``.

        .. seealso::
           :meth:`cupy.ndarray.sum`, :meth:`numpy.ndarray.sum`
        """
        if dtype is None:
            return self.__cupy_override_reduction_kernel__(
                _math.sum_auto_dtype, axis, dtype, out, keepdims)
        else:
            return self.__cupy_override_reduction_kernel__(
                _math.sum_keep_dtype, axis, dtype, out, keepdims)

    def swapaxes(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support swapaxes.')

    def take(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support take.')

    def toDlpack(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support toDlpack.')

    def tobytes(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support tobytes.')

    def tofile(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support tofile.')

    def tolist(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support tolist.')

    def trace(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support trace.')

    def transpose(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support transpose.')

    def var(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support var.')

    def view(self, *args, **kwargs):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support view.')

    @property
    def T(self):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support T.')

    @property
    def base(self):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support base.')

    @property
    def cstruct(self):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support cstruct.')

    @property
    def data(self):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support data.')

    @property
    def device(self):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support device.')

    @property
    def flags(self):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support flags.')

    @property
    def flat(self):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support flat.')

    @property
    def imag(self):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support imag.')

    @property
    def real(self):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support real.')

    @property
    def shape(self):
        """Tuple of array dimensions.

        Assignment to this property is currently not supported.

        .. seealso: :attr:`cupy.ndarray.shape`, :attr:`numpy.ndarray.shape`

        """
        return super().shape

    @shape.setter
    def shape(self, newshape):
        raise NotImplementedError(
            'DistributedArray currently does not support assignment to shape.')

    @property
    def strides(self):
        """Not supported."""
        raise NotImplementedError(
            'DistributedArray currently does not support strides.')


def distributed_array(
    array: ArrayLike,
    index_map: dict[int, Any],
    mode: _modes.Mode = _modes.REPLICA,
) -> DistributedArray:
    """Creates a distributed array from the given data.

    This function does not check if all elements of the given array are stored
    in some of the chunks.

    Args:
        array (array_like): :class:`DistributedArray` object,
            :class:`cupy.ndarray` object or any other object that can be passed
            to :func:`numpy.array`.
        index_map (dict from int to array indices): Indices for the chunks
            that devices with designated IDs own. One device can have multiple
            chunks, which can be specified as a list of array indices.
        mode (mode object, optional): Mode that determines how overlaps
            of the chunks are interpreted. Defaults to
            ``cupyx.distributed.array.REPLICA``.

    .. seealso::
            :attr:`DistributedArray.mode` for details about modes.

    Example:
        >>> array = cupy.arange(9).reshape(3, 3)
        >>> A = distributed_array(
        ...     array,
        ...     {0: [(slice(2), slice(2)),  # array[:2, :2]
        ...          slice(None, None, 2)], # array[::2]
        ...      1:  (slice(1, None), 2)})  # array[1:, 2]
    """
    if isinstance(array, DistributedArray):
        if array.mode != mode:
            array = array.change_mode(mode)
        if array.index_map != index_map:
            array = array.reshard(index_map)
        return DistributedArray(
            array.shape, array.dtype, array._chunks_map, array._mode,
            array._comms)

    if isinstance(array, (numpy.ndarray, ndarray)):
        if mode != _modes.REPLICA:
            array = array.copy()
    else:
        array = numpy.array(array)

    index_map = _index_arith._normalize_index_map(array.shape, index_map)
    comms = None

    # Define how to form a chunk from (dev, idx, src_array)
    make_chunk: Callable[
        [int, int, tuple[slice, ...], ndarray, Optional[list[Any]]],
        _Chunk
    ]

    if isinstance(array, ndarray):
        src_dev = array.device.id
        devices = index_map.keys() | {array.device.id}
        comms = _data_transfer._create_communicators(devices)
        make_chunk = _make_chunk_async
    else:
        src_dev = -1
        make_chunk = _make_chunk_sync

    chunks_map: dict[int, list[_Chunk]] = {}
    for dev, idxs in index_map.items():
        chunks_map[dev] = []

        for idx in idxs:
            chunk_array = array[idx]
            chunk = make_chunk(src_dev, dev, idx, chunk_array, comms)
            chunks_map[dev].append(chunk)
            if (mode is not _modes.REPLICA
                    and not mode.idempotent):
                array[idx] = mode.identity_of(array.dtype)

    return DistributedArray(
        array.shape, array.dtype, chunks_map, mode, comms)



import unittest

import numpy

import cupy
from cupy import testing
import cupyx.scipy.special  # NOQA


def _boundary_inputs(boundary, rtol, atol):
    left = boundary * (1 - numpy.copysign(rtol, boundary)) - atol
    right = boundary * (1 + numpy.copysign(rtol, boundary)) + atol
    return [left, boundary, right]


@testing.with_requires('scipy')
class _TestBase:

    @testing.with_requires('scipy<1.16.0')
    def test_erf(self):
        self.check_unary('erf')

    @testing.with_requires('scipy<1.16.0')
    def test_erfc(self):
        self.check_unary('erfc')

    @testing.with_requires('scipy<1.16.0')
    def test_erfcx(self):
        self.check_unary('erfcx')

    def test_erfinv(self):
        self.check_unary('erfinv')
        self.check_unary_random('erfinv', scale=2, offset=-1)
        self.check_unary_boundary('erfinv', boundary=-1)
        self.check_unary_boundary('erfinv', boundary=1)

    def test_erfcinv(self):
        self.check_unary('erfcinv')
        self.check_unary_random('erfcinv', scale=2, offset=0)
        self.check_unary_boundary('erfcinv', boundary=0)
        self.check_unary_boundary('erfcinv', boundary=2)


@testing.with_requires('scipy')
class TestSpecial(unittest.TestCase, _TestBase):

    @testing.for_dtypes(['e', 'f', 'd'])
    @testing.numpy_cupy_allclose(atol=1e-5, scipy_name='scp')
    def check_unary(self, name, xp, scp, dtype):
        import scipy.special  # NOQA

        a = testing.shaped_arange((2, 3), xp, dtype)
        return getattr(scp.special, name)(a)

    @testing.for_dtypes(['f', 'd'])
    @testing.numpy_cupy_allclose(atol=1e-5, scipy_name='scp')
    def check_unary_random(self, name, xp, scp, dtype, scale, offset):
        import scipy.special  # NOQA

        a = testing.shaped_random((2, 3), xp, dtype, scale=scale) + offset
        return getattr(scp.special, name)(a)

    @testing.for_dtypes(['f', 'd'])
    @testing.numpy_cupy_allclose(atol=1e-5, scipy_name='scp')
    def check_unary_boundary(self, name, xp, scp, dtype, boundary):
        import scipy.special  # NOQA

        a = _boundary_inputs(boundary, 1.0 / 1024, 1.0 / 1024)
        a = xp.array(a, dtype=dtype)
        return getattr(scp.special, name)(a)

    @testing.with_requires('scipy>=1.4.0')
    @testing.for_dtypes(['f', 'd'])
    def test_erfinv_behavior(self, dtype):
        a = cupy.empty((1,), dtype=dtype)

        a[:] = 1.0 + 1E-6
        a = cupyx.scipy.special.erfinv(a)
        assert cupy.isnan(a)
        a[:] = -1.0 - 1E-6
        a = cupyx.scipy.special.erfinv(a)
        assert cupy.isnan(a)
        a[:] = 1.0
        a = cupyx.scipy.special.erfinv(a)
        assert numpy.isposinf(cupy.asnumpy(a))
        a[:] = -1.0
        a = cupyx.scipy.special.erfinv(a)
        assert numpy.isneginf(cupy.asnumpy(a))

    @testing.with_requires('scipy>=1.4.0')
    @testing.for_dtypes(['f', 'd'])
    def test_erfcinv_behavior(self, dtype):
        a = cupy.empty((1,), dtype=dtype)

        a[:] = 2.0 + 1E-6
        a = cupyx.scipy.special.erfcinv(a)
        assert cupy.isnan(a)
        a[:] = 0.0 - 1E-6
        a = cupyx.scipy.special.erfcinv(a)
        assert cupy.isnan(a)
        a[:] = 0.0
        a = cupyx.scipy.special.erfcinv(a)
        assert numpy.isposinf(cupy.asnumpy(a))
        a[:] = 2.0
        a = cupyx.scipy.special.erfcinv(a)
        assert numpy.isneginf(cupy.asnumpy(a))


@testing.with_requires('scipy')
class TestFusionSpecial(unittest.TestCase, _TestBase):

    @testing.for_dtypes(['e', 'f', 'd'])
    @testing.numpy_cupy_allclose(atol=1e-5, scipy_name='scp')
    def check_unary(self, name, xp, scp, dtype):
        import scipy.special  # NOQA

        a = testing.shaped_arange((2, 3), xp, dtype)

        @cupy.fuse()
        def f(x):
            return getattr(scp.special, name)(x)

        return f(a)

    @testing.for_dtypes(['f', 'd'])
    @testing.numpy_cupy_allclose(atol=1e-5, scipy_name='scp')
    def check_unary_random(self, name, xp, scp, dtype, scale, offset):
        import scipy.special  # NOQA

        a = testing.shaped_random((2, 3), xp, dtype, scale=scale) + offset

        @cupy.fuse()
        def f(x):
            return getattr(scp.special, name)(x)

        return f(a)

    @testing.for_dtypes(['f', 'd'])
    @testing.numpy_cupy_allclose(atol=1e-5, scipy_name='scp')
    def check_unary_boundary(self, name, xp, scp, dtype, boundary):
        import scipy.special  # NOQA

        a = _boundary_inputs(boundary, 1.0 / 1024, 1.0 / 1024)
        a = xp.array(a, dtype=dtype)

        @cupy.fuse()
        def f(x):
            return getattr(scp.special, name)(x)

        return f(a)



import unittest
from unittest import mock

from cupy import cuda
from cupy import prof


@unittest.skipUnless(cuda.nvtx.available, 'nvtx is required for time_range')
class TestTimeRange(unittest.TestCase):

    def test_time_range(self):
        push_patch = mock.patch('cupy.cuda.nvtx.RangePush')
        pop_patch = mock.patch('cupy.cuda.nvtx.RangePop')
        with push_patch as push, pop_patch as pop:
            with prof.time_range('test:time_range', color_id=-1):
                pass
            push.assert_called_once_with('test:time_range', -1)
            pop.assert_called_once_with()

    def test_time_range_with_ARGB(self):
        push_patch = mock.patch('cupy.cuda.nvtx.RangePushC')
        pop_patch = mock.patch('cupy.cuda.nvtx.RangePop')
        with push_patch as push, pop_patch as pop:
            with prof.time_range('test:time_range_with_ARGB',
                                 argb_color=0xFF00FF00):
                pass
            push.assert_called_once_with(
                'test:time_range_with_ARGB', 0xFF00FF00)
            pop.assert_called_once_with()

    def test_time_range_err(self):
        push_patch = mock.patch('cupy.cuda.nvtx.RangePush')
        pop_patch = mock.patch('cupy.cuda.nvtx.RangePop')
        with push_patch as push, pop_patch as pop:
            try:
                with prof.time_range('test:time_range_error', -1):
                    raise Exception()
            except Exception:
                pass
            push.assert_called_once_with('test:time_range_error', -1)
            pop.assert_called_once_with()

    def test_TimeRangeDecorator(self):
        push_patch = mock.patch('cupy.cuda.nvtx.RangePush')
        pop_patch = mock.patch('cupy.cuda.nvtx.RangePop')
        with push_patch as push, pop_patch as pop:
            @prof.TimeRangeDecorator()
            def f():
                pass
            f()
            # Default value of color id is -1
            push.assert_called_once_with('f', -1)
            pop.assert_called_once_with()

    def test_TimeRangeDecorator_with_ARGB(self):
        push_patch = mock.patch('cupy.cuda.nvtx.RangePushC')
        pop_patch = mock.patch('cupy.cuda.nvtx.RangePop')
        with push_patch as push, pop_patch as pop:
            @prof.TimeRangeDecorator(argb_color=0xFFFF0000)
            def f():
                pass
            f()
            push.assert_called_once_with('f', 0xFFFF0000)
            pop.assert_called_once_with()

    def test_TimeRangeDecorator_err(self):
        push_patch = mock.patch('cupy.cuda.nvtx.RangePush')
        pop_patch = mock.patch('cupy.cuda.nvtx.RangePop')
        with push_patch as push, pop_patch as pop:
            @prof.TimeRangeDecorator()
            def f():
                raise Exception()
            try:
                f()
            except Exception:
                pass
            # Default value of color id is -1
            push.assert_called_once_with('f', -1)
            pop.assert_called_once_with()


class TestTimeRangeNVTXUnavailable(unittest.TestCase):

    def setUp(self):
        self.nvtx_available = cuda.nvtx.available
        cuda.nvtx.available = False

    def tearDown(self):
        cuda.nvtx.available = self.nvtx_available

    def test_time_range(self):
        with self.assertRaises(RuntimeError):
            with prof.time_range(''):
                pass

    def test_time_range_decorator(self):
        with self.assertRaises(RuntimeError):
            prof.TimeRangeDecorator()



cpdef enum:
    # need to revisit this when cython supports C++ enums (in 3.0)
    # https://stackoverflow.com/a/67138945

    cudaMemoryTypeHost = 1
    cudaMemoryTypeDevice = 2

    cudaIpcMemLazyEnablePeerAccess = 1

    cudaMemAttachGlobal = 1
    cudaMemAttachHost = 2
    cudaMemAttachSingle = 4

    cudaCpuDeviceId = -1
    cudaInvalidDeviceId = -2

    cudaMemAdviseSetReadMostly = 1
    cudaMemAdviseUnsetReadMostly = 2
    cudaMemAdviseSetPreferredLocation = 3
    cudaMemAdviseUnsetPreferredLocation = 4
    cudaMemAdviseSetAccessedBy = 5
    cudaMemAdviseUnsetAccessedBy = 6

    CUDA_R_32F = 0  # 32 bit real
    CUDA_R_64F = 1  # 64 bit real
    CUDA_R_16F = 2  # 16 bit real
    CUDA_R_8I = 3  # 8 bit real as a signed integer
    CUDA_C_32F = 4  # 32 bit complex
    CUDA_C_64F = 5  # 64 bit complex
    CUDA_C_16F = 6  # 16 bit complex
    CUDA_C_8I = 7  # 8 bit complex as a pair of signed integers
    CUDA_R_8U = 8  # 8 bit real as a signed integer
    CUDA_C_8U = 9  # 8 bit complex as a pair of signed integers

    # CUDA Limits
    cudaLimitStackSize = 0x00
    cudaLimitPrintfFifoSize = 0x01
    cudaLimitMallocHeapSize = 0x02
    cudaLimitDevRuntimeSyncDepth = 0x03
    cudaLimitDevRuntimePendingLaunchCount = 0x04
    cudaLimitMaxL2FetchGranularity = 0x05

    # cudaChannelFormatKind
    cudaChannelFormatKindSigned = 0
    cudaChannelFormatKindUnsigned = 1
    cudaChannelFormatKindFloat = 2
    cudaChannelFormatKindNone = 3

    # CUDA array flags
    cudaArrayDefault = 0
    # cudaArrayLayered = 1
    cudaArraySurfaceLoadStore = 2
    # cudaArrayCubemap = 4
    # cudaArrayTextureGather = 8

    # cudaResourceType
    cudaResourceTypeArray = 0
    cudaResourceTypeMipmappedArray = 1
    cudaResourceTypeLinear = 2
    cudaResourceTypePitch2D = 3

    # cudaTextureAddressMode
    cudaAddressModeWrap = 0
    cudaAddressModeClamp = 1
    cudaAddressModeMirror = 2
    cudaAddressModeBorder = 3

    # cudaTextureFilterMode
    cudaFilterModePoint = 0
    cudaFilterModeLinear = 1

    # cudaTextureReadMode
    cudaReadModeElementType = 0
    cudaReadModeNormalizedFloat = 1

    # cudaMemPoolAttr
    # ----- added since 11.2 -----
    cudaMemPoolReuseFollowEventDependencies = 0x1
    cudaMemPoolReuseAllowOpportunistic = 0x2
    cudaMemPoolReuseAllowInternalDependencies = 0x3
    cudaMemPoolAttrReleaseThreshold = 0x4
    # ----- added since 11.3 -----
    cudaMemPoolAttrReservedMemCurrent = 0x5
    cudaMemPoolAttrReservedMemHigh = 0x6
    cudaMemPoolAttrUsedMemCurrent = 0x7
    cudaMemPoolAttrUsedMemHigh = 0x8

    # cudaMemAllocationType
    cudaMemAllocationTypePinned = 0x1

    # cudaMemAllocationHandleType
    cudaMemHandleTypeNone = 0x0
    cudaMemHandleTypePosixFileDescriptor = 0x1
    # cudaMemHandleTypeWin32 = 0x2
    # cudaMemHandleTypeWin32Kmt = 0x4

    # cudaMemLocationType
    cudaMemLocationTypeInvalid = 0          # CUDA 12.0
    cudaMemLocationTypeDevice = 1
    cudaMemLocationTypeHost = 2             # CUDA 12.0
    cudaMemLocationTypeHostNuma = 3         # CUDA 12.0
    cudaMemLocationTypeHostNumaCurrent = 4  # CUDA 12.0


# This was a legacy mistake: the prefix "cuda" should have been removed
# so that we can directly assign their C counterparts here. Now because
# of backward compatibility and no flexible Cython macro (IF/ELSE), we
# have to duplicate the enum. (CUDA and HIP use different values!)
IF CUPY_HIP_VERSION > 0:
    # separate in groups of 10 for easier counting...
    cpdef enum:
        cudaDevAttrMaxThreadsPerBlock = 0
        cudaDevAttrMaxBlockDimX
        cudaDevAttrMaxBlockDimY
        cudaDevAttrMaxBlockDimZ
        cudaDevAttrMaxGridDimX
        cudaDevAttrMaxGridDimY
        cudaDevAttrMaxGridDimZ
        cudaDevAttrMaxSharedMemoryPerBlock
        cudaDevAttrTotalConstantMemory
        cudaDevAttrWarpSize

        cudaDevAttrMaxRegistersPerBlock
        cudaDevAttrClockRate
        cudaDevAttrMemoryClockRate
        cudaDevAttrGlobalMemoryBusWidth
        cudaDevAttrMultiProcessorCount
        cudaDevAttrComputeMode
        cudaDevAttrL2CacheSize
        cudaDevAttrMaxThreadsPerMultiProcessor
        # The following are exposed as "deviceAttributeCo..."
        # cudaDevAttrComputeCapabilityMajor
        # cudaDevAttrComputeCapabilityMinor

        cudaDevAttrConcurrentKernels = 20
        cudaDevAttrPciBusId
        cudaDevAttrPciDeviceId
        cudaDevAttrMaxSharedMemoryPerMultiprocessor
        cudaDevAttrIsMultiGpuBoard
        cudaDevAttrIntegrated
        cudaDevAttrCooperativeLaunch
        cudaDevAttrCooperativeMultiDeviceLaunch
        cudaDevAttrMaxTexture1DWidth
        cudaDevAttrMaxTexture2DWidth

        cudaDevAttrMaxTexture2DHeight
        cudaDevAttrMaxTexture3DWidth
        cudaDevAttrMaxTexture3DHeight
        cudaDevAttrMaxTexture3DDepth
        # The following attributes do not exist in CUDA and cause segfaults
        # if we try to access them
        # hipDeviceAttributeHdpMemFlushCntl
        # hipDeviceAttributeHdpRegFlushCntl
        cudaDevAttrMaxPitch = 36
        cudaDevAttrTextureAlignment
        cudaDevAttrTexturePitchAlignment
        cudaDevAttrKernelExecTimeout

        cudaDevAttrCanMapHostMemory
        cudaDevAttrEccEnabled
        cudaDevAttrMemoryPoolsSupported = 0
        # The following attributes do not exist in CUDA
        # hipDeviceAttributeCooperativeMultiDeviceUnmatchedFunc
        # hipDeviceAttributeCooperativeMultiDeviceUnmatchedGridDim
        # hipDeviceAttributeCooperativeMultiDeviceUnmatchedBlockDim
        # hipDeviceAttributeCooperativeMultiDeviceUnmatchedSharedMem

        # The rest do not have HIP correspondence...
        # TODO(leofang): should we expose them anyway, with a value -1 to
        # indicate they cannot be used in HIP?
        # cudaDevAttrGpuOverlap
        # cudaDevAttrMaxTexture2DLayeredWidth
        # cudaDevAttrMaxTexture2DLayeredHeight
        # cudaDevAttrMaxTexture2DLayeredLayers
        # cudaDevAttrSurfaceAlignment
        # cudaDevAttrTccDriver
        # cudaDevAttrAsyncEngineCount
        # cudaDevAttrUnifiedAddressing
        # cudaDevAttrMaxTexture1DLayeredWidth
        # cudaDevAttrMaxTexture1DLayeredLayers
        # cudaDevAttrMaxTexture2DGatherWidth
        # cudaDevAttrMaxTexture2DGatherHeight
        # cudaDevAttrMaxTexture3DWidthAlt
        # cudaDevAttrMaxTexture3DHeightAlt
        # cudaDevAttrMaxTexture3DDepthAlt
        # cudaDevAttrPciDomainId
        # cudaDevAttrMaxTextureCubemapWidth
        # cudaDevAttrMaxTextureCubemapLayeredWidth
        # cudaDevAttrMaxTextureCubemapLayeredLayers
        # cudaDevAttrMaxSurface1DWidth
        # cudaDevAttrMaxSurface2DWidth
        # cudaDevAttrMaxSurface2DHeight
        # cudaDevAttrMaxSurface3DWidth
        # cudaDevAttrMaxSurface3DHeight
        # cudaDevAttrMaxSurface3DDepth
        # cudaDevAttrMaxSurface1DLayeredWidth
        # cudaDevAttrMaxSurface1DLayeredLayers
        # cudaDevAttrMaxSurface2DLayeredWidth
        # cudaDevAttrMaxSurface2DLayeredHeight
        # cudaDevAttrMaxSurface2DLayeredLayers
        # cudaDevAttrMaxSurfaceCubemapWidth
        # cudaDevAttrMaxSurfaceCubemapLayeredWidth
        # cudaDevAttrMaxSurfaceCubemapLayeredLayers
        # cudaDevAttrMaxTexture1DLinearWidth
        # cudaDevAttrMaxTexture2DLinearWidth
        # cudaDevAttrMaxTexture2DLinearHeight
        # cudaDevAttrMaxTexture2DLinearPitch
        # cudaDevAttrMaxTexture2DMipmappedWidth
        # cudaDevAttrMaxTexture2DMipmappedHeight
        # cudaDevAttrMaxTexture1DMipmappedWidth
        # cudaDevAttrStreamPrioritiesSupported
        # cudaDevAttrGlobalL1CacheSupported
        # cudaDevAttrLocalL1CacheSupported
        # cudaDevAttrMaxRegistersPerMultiprocessor
        # cudaDevAttrMultiGpuBoardGroupID
        # cudaDevAttrHostNativeAtomicSupported
        # cudaDevAttrSingleToDoublePrecisionPerfRatio
        # cudaDevAttrComputePreemptionSupported
        # cudaDevAttrCanUseHostPointerForRegisteredMem
        # cudaDevAttrReserved92
        # cudaDevAttrReserved93
        # cudaDevAttrReserved94
        # cudaDevAttrMaxSharedMemoryPerBlockOptin
        # cudaDevAttrCanFlushRemoteWrites
        # cudaDevAttrHostRegisterSupported
    IF CUPY_HIP_VERSION >= 310:
        cpdef enum:
            # hipDeviceAttributeAsicRevision  # does not exist in CUDA
            cudaDevAttrManagedMemory = 47
            cudaDevAttrDirectManagedMemAccessFromHost
            cudaDevAttrConcurrentManagedAccess

            cudaDevAttrPageableMemoryAccess
            cudaDevAttrPageableMemoryAccessUsesHostPageTables
ELSE:
    # For CUDA/RTD
    cpdef enum:
        cudaDevAttrMaxThreadsPerBlock = 1
        cudaDevAttrMaxBlockDimX
        cudaDevAttrMaxBlockDimY
        cudaDevAttrMaxBlockDimZ
        cudaDevAttrMaxGridDimX
        cudaDevAttrMaxGridDimY
        cudaDevAttrMaxGridDimZ
        cudaDevAttrMaxSharedMemoryPerBlock
        cudaDevAttrTotalConstantMemory
        cudaDevAttrWarpSize
        cudaDevAttrMaxPitch
        cudaDevAttrMaxRegistersPerBlock
        cudaDevAttrClockRate
        cudaDevAttrTextureAlignment
        cudaDevAttrGpuOverlap
        cudaDevAttrMultiProcessorCount
        cudaDevAttrKernelExecTimeout
        cudaDevAttrIntegrated
        cudaDevAttrCanMapHostMemory
        cudaDevAttrComputeMode
        cudaDevAttrMaxTexture1DWidth
        cudaDevAttrMaxTexture2DWidth
        cudaDevAttrMaxTexture2DHeight
        cudaDevAttrMaxTexture3DWidth
        cudaDevAttrMaxTexture3DHeight
        cudaDevAttrMaxTexture3DDepth
        cudaDevAttrMaxTexture2DLayeredWidth
        cudaDevAttrMaxTexture2DLayeredHeight
        cudaDevAttrMaxTexture2DLayeredLayers
        cudaDevAttrSurfaceAlignment
        cudaDevAttrConcurrentKernels
        cudaDevAttrEccEnabled
        cudaDevAttrPciBusId
        cudaDevAttrPciDeviceId
        cudaDevAttrTccDriver
        cudaDevAttrMemoryClockRate
        cudaDevAttrGlobalMemoryBusWidth
        cudaDevAttrL2CacheSize
        cudaDevAttrMaxThreadsPerMultiProcessor
        cudaDevAttrAsyncEngineCount
        cudaDevAttrUnifiedAddressing
        cudaDevAttrMaxTexture1DLayeredWidth
        cudaDevAttrMaxTexture1DLayeredLayers  # = 43; 44 is missing
        cudaDevAttrMaxTexture2DGatherWidth = 45
        cudaDevAttrMaxTexture2DGatherHeight
        cudaDevAttrMaxTexture3DWidthAlt
        cudaDevAttrMaxTexture3DHeightAlt
        cudaDevAttrMaxTexture3DDepthAlt
        cudaDevAttrPciDomainId
        cudaDevAttrTexturePitchAlignment
        cudaDevAttrMaxTextureCubemapWidth
        cudaDevAttrMaxTextureCubemapLayeredWidth
        cudaDevAttrMaxTextureCubemapLayeredLayers
        cudaDevAttrMaxSurface1DWidth
        cudaDevAttrMaxSurface2DWidth
        cudaDevAttrMaxSurface2DHeight
        cudaDevAttrMaxSurface3DWidth
        cudaDevAttrMaxSurface3DHeight
        cudaDevAttrMaxSurface3DDepth
        cudaDevAttrMaxSurface1DLayeredWidth
        cudaDevAttrMaxSurface1DLayeredLayers
        cudaDevAttrMaxSurface2DLayeredWidth
        cudaDevAttrMaxSurface2DLayeredHeight
        cudaDevAttrMaxSurface2DLayeredLayers
        cudaDevAttrMaxSurfaceCubemapWidth
        cudaDevAttrMaxSurfaceCubemapLayeredWidth
        cudaDevAttrMaxSurfaceCubemapLayeredLayers
        cudaDevAttrMaxTexture1DLinearWidth
        cudaDevAttrMaxTexture2DLinearWidth
        cudaDevAttrMaxTexture2DLinearHeight
        cudaDevAttrMaxTexture2DLinearPitch
        cudaDevAttrMaxTexture2DMipmappedWidth
        cudaDevAttrMaxTexture2DMipmappedHeight
        # The following are exposed as "deviceAttributeCo..."
        # cudaDevAttrComputeCapabilityMajor  # = 75
        # cudaDevAttrComputeCapabilityMinor  # = 76
        cudaDevAttrMaxTexture1DMipmappedWidth = 77
        cudaDevAttrStreamPrioritiesSupported
        cudaDevAttrGlobalL1CacheSupported
        cudaDevAttrLocalL1CacheSupported
        cudaDevAttrMaxSharedMemoryPerMultiprocessor
        cudaDevAttrMaxRegistersPerMultiprocessor
        cudaDevAttrManagedMemory
        cudaDevAttrIsMultiGpuBoard
        cudaDevAttrMultiGpuBoardGroupID
        cudaDevAttrHostNativeAtomicSupported
        cudaDevAttrSingleToDoublePrecisionPerfRatio
        cudaDevAttrPageableMemoryAccess
        cudaDevAttrConcurrentManagedAccess
        cudaDevAttrComputePreemptionSupported
        cudaDevAttrCanUseHostPointerForRegisteredMem
        cudaDevAttrReserved92
        cudaDevAttrReserved93
        cudaDevAttrReserved94
        cudaDevAttrCooperativeLaunch
        cudaDevAttrCooperativeMultiDeviceLaunch
        cudaDevAttrMaxSharedMemoryPerBlockOptin
        cudaDevAttrCanFlushRemoteWrites
        cudaDevAttrHostRegisterSupported
        cudaDevAttrPageableMemoryAccessUsesHostPageTables
        cudaDevAttrDirectManagedMemAccessFromHost  # = 101
        # added since CUDA 11.0
        cudaDevAttrMaxBlocksPerMultiprocessor = 106
        cudaDevAttrReservedSharedMemoryPerBlock = 111
        # added since CUDA 11.1
        cudaDevAttrSparseCudaArraySupported = 112
        cudaDevAttrHostRegisterReadOnlySupported = 113
        # added since CUDA 11.2
        cudaDevAttrMaxTimelineSemaphoreInteropSupported = 114
        cudaDevAttrMemoryPoolsSupported = 115
        # added since CUDA 11.3
        cudaDevAttrGPUDirectRDMASupported
        cudaDevAttrGPUDirectRDMAFlushWritesOptions
        cudaDevAttrGPUDirectRDMAWritesOrdering
        cudaDevAttrMemoryPoolSupportedHandleTypes
